# -*- coding: utf-8 -*-
# IMPACT-CALCULATOR
# from a given Jacobian matrix it calculates:
# The response to small perturbations I.
# The sensitivity of each variable to perturbations
# The influence of each variable through perturbations
# Please see our paper for details.
####################################################################
#(C) BY H.AUFDERHEIDE 
# THIS SCRIPT IS DISTRIBUTED AS-IS
# YOU USE IT AT YOUR OWN RISK
#
## This work is licensed under the Creative Commons Attribution-NonCommercial 3.0 Unported License.
# To view a copy of this license, visit http://creativecommons.org/licenses/by-nc/3.0/ or
# send a letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.
#
# The author takes no responsability whatshowever for the correctness and correct function of this script
# and is not liable for any damage caused through its correct or incorrect use.
######################################################################
# USAGE: Run in a terminal:
# python Impact.py <Jacobian> <Perturbation>
# <Jacobian> denotes the name of a text file that contains the Jacobian matrix, columns are separated by tabulators.
# <Perturbation> (optional) is a file that contains the perturbation matrix, columns separated by tabulators.
#######################################################################

import numpy as np
import sys

################################HELPERS#################################
 
def OpenJacobian():		
#Loads a Jacobian matrix from the file given as first argument
    if (len(sys.argv)<2):  sys.exit("No File with Jacobian specified.")
    J=np.loadtxt(sys.argv[1],delimiter='\t',comments='#')
    if (len(J)!=len(J[0])):  sys.exit("Jacobian must be a square matrix.")	#A very basic check if we have a square matrix
    return J
    
def CalculateSensitivity(J):  
#Calculates the Sensitivity-Vector from the Jacobian
    EVals,EVecs=np.linalg.eig(J)
    Sens=np.zeros(len(EVecs))
    
    for i_ev,ev in enumerate(EVals):
      if np.real(ev)>0:  print "WARNING: System is not at a stable stady state. "+str(ev)+" has a positive real part."
      if (np.abs(ev)< 0.00000001) :  print "WARNING: System is very close to a bifurcation."
      if (np.real(ev)==0) :  
	print "WARNING: System is at a bifurcation. Returning first encountered neutral direction."
        Sens=EVecs[:,i_ev]
	break
      Sens=Sens+abs(EVecs[:,i_ev])*np.real(1./ev)
    return np.log(Sens)

def CalculateInfluence(J):		
#Calculates the Influence-Vector of all nodes from the Jacobian (the transposed Jacobian generates the left-side spectrum)
    return CalculateSensitivity(J.T)	



def OpenPerturbation():
#Loads a matrix containing in each column j one perturbation of species i at the steady state K_i,j=dA_i/dA_j
    if (len(sys.argv)<3):  sys.exit("No File for Perturbation Specified.")
    K=np.loadtxt(sys.argv[2],delimiter='\t',comments='#')
    return K

def CalculateImpact(J,K):
#Calculates the Impact as the solution to J*Impact=K
    if (len(J)!=len(K)):  sys.exit("Perturbation and Jacobian describe systems with different number of nodes.")
    Impact=-np.linalg.solve(J, K)
    return Impact


###############################MAIN PROGRAM########################################

#READ JACOBIAN
Jacobian=OpenJacobian() 

#CALCULATE SENSITIVITY AND INFLUENCE
Sensitivity=CalculateSensitivity(Jacobian) 
Influence=CalculateInfluence(Jacobian)

#SAVE SENSITIVITY AND INFLUENCE
np.savetxt("SensitivityAndInfluence-"+sys.argv[1],zip(Sensitivity,Influence),delimiter="\t",fmt="%f") 

#CALCULATE IMPACT ( if perturbation file is given. )
if (len(sys.argv)>2):
  Perturbation=OpenPerturbation()
  Impact=CalculateImpact(Jacobian,Perturbation)
#SAVE THE IMPACT  
  np.savetxt("Response-"+sys.argv[2],Impact,delimiter="\t",fmt="%f")
